# -*- coding: utf-8 -*-
verbose=True
# Suppress SMART warnings (True/False)
report_warnings=True
# Modify to your needs:
statefiledir='/var/lib/munin/plugin-state/'
# You may not modify anything below this line

import os, sys, string, pickle,re
from math import log
plugin_version="2.0"
smart_values ={}
def verboselog(s):
    sys.stderr.write("munin_smart"+': '+s+'\n')

if not verbose :
    verboselog = lambda s: None

def read_values(hard_drive):
  global smart_values
  
  try :
    verboselog('Reading S.M.A.R.T values')
    os.putenv('LC_ALL','C')
    if(hard_drive.find("twa")<0):
      smart_command = os.getenv('smartpath','/usr/sbin/smartctl')+' '+os.getenv('smartargs','-a')+(os.getenv('ignorestandby',False) and ' ' or ' -n standby ')+'-A -i /dev/'+hard_drive;
      
    else:
      hard_drive = hard_drive.split("_");
      controller = hard_drive[0];
      port = hard_drive[1]
      smart_command = os.getenv('smartpath','/usr/sbin/smartctl')+' '+os.getenv('smartargs','-a')+(os.getenv('ignorestandby',False) and ' ' or ' -n standby ')+'-H -A -i -c -l error -l selftest -l selective -a -A -d 3ware,'+port+' /dev/'+controller;
    verboselog(smart_command);  
    smart_output=os.popen(smart_command)
    read_values=0
    for l in smart_output :
      if l[:-1]=='' :
        read_values=0
      elif l[:13]=='Device Model:' or l[:7]=='Device:' :
        model_list=string.split(string.split(l,':')[1])
        try: model_list.remove('Version')
        except : None
        model=string.join(model_list)
      if read_values==1 :
        smart_attribute=string.split(l)
        smart_values[string.replace(smart_attribute[1],'-','_')]={"value":smart_attribute[3],"threshold":smart_attribute[5]}
      elif l[:18]=="ID# ATTRIBUTE_NAME" :
        # Start reading the Attributes block
        read_values=1
    exit_status=smart_output.close()
    if exit_status!=None :
      # smartctl exit code is a bitmask, check man page.
      num_exit_status=int(exit_status/256) # Python convention
      if int(log(num_exit_status,2))<=2 :  # bit code
        verboselog('smartctl cannot access S.M.A.R.T values on drive '+hard_drive+'. Command exited with code '+str(num_exit_status)+' (bit '+str(int(log(num_exit_status,2)))+')')
      else :
        verboselog('smartctl exited with code '+str(num_exit_status)+' (bit '+str(int(log(num_exit_status,2)))+'). '+hard_drive+' may be FAILING RIGHT NOW!')
    else :
      num_exit_status=0
  except :
      verboselog('Cannot access S.M.A.R.T values! Check user rights or proper smartmontools installation/arguments.')
      sys.exit(1)
  if smart_values=={} :
    verboselog('Can\'t find any S.M.A.R.T values in smartctl output!')
    sys.exit(1)
  smart_values["smartctl_exit_status"]={"value":str(num_exit_status),"threshold":"1"}
  try : smart_values["model"]=model
  # For some reason we may have no value for "model"
  except : smart_values["model"]="unknown"
  return(exit_status)

def find_smart_drives() :
  # Try to autodetect Linux, *BSD, SunOS drives. Don't try to autodetect drives on a 3Ware card.
  drives=[]
  if os.uname()[0]=="Linux" :
    if os.path.exists('/sys/block/'):
      # Running 2.6
      try :
        for drive in os.listdir('/sys/block/') :
          if drive[:2] in ['md','fd','lo','ra','dm'] : continue # Ignore MD, Floppy, loop , RAM and LVM devices.
          try :
            verboselog('Trying '+drive+'...')
            exit_status=read_values(drive)
            if (exit_status==None or int(log(int(exit_status/256),2))>2):
              verboselog('Append '+drive+'...')
              drives.append(drive);
          except :
            verboselog('Ignore '+drive+'...')
            continue
      except : 
        verboselog('Failed to list devices in /sys/block')
    else :
      verboselog('Not running linux2.6, failing back to /proc/partitions')
      try :
        partitions=open('/proc/partitions','r')
        L=partitions.readlines()
        for l in L :
          words=string.split(l)
          if len(words)==0 or words[0][0] not in string.digits : continue
          if words[0] in ['1','9','58','254'] : continue # Ignore RAM, md, LVM and LVM2 devices
          if words[-1][-1] not in string.digits : 
            try :
              verboselog('Trying '+words[-1]+'...')
              exit_status=read_values(words[-1])
              if (exit_status==None or int(log(int(exit_status/256),2))>2):
                drives.append(words[-1])
            except :
              continue
          verboselog('Found drives in /proc/partitions ! '+str(drives))
      except :
        verboselog('Failed to list devices in /proc/partitions')
    if(os.path.exists('/usr/sbin/tw_cli')):
      detectObject = re.compile("^[cp]\\d+")
      twi_list_output = os.popen('/usr/sbin/tw_cli show');
      for controllerLine in twi_list_output:
        controller = detectObject.search(controllerLine);
        if(controller is not None): 
          controller = controller.group();
          twi_controller_output = os.popen('/usr/sbin/tw_cli /'+controller+' show');
          for portLine in twi_controller_output:
            port = detectObject.search(portLine);
            if(port is not None):
              port = port.group();
              drive = controller.replace("c","twa")+"_"+port.replace("p","")
              drives.append(drive);
  elif os.uname()[0]=="OpenBSD" :
      try :
          sysctl_kerndisks=os.popen('sysctl hw.disknames')
          kerndisks=string.strip(sysctl_kerndisks.readline())
          for drive in string.split(kerndisks[string.rindex(kerndisks,'=')+1:],',') :
              if drive[:2] in ['md','cd','fd'] : continue # Ignore Memory Disks, CD-ROM drives and Floppy
              try :
                  verboselog('Trying '+drive+'c...')
                  exit_status=read_values(drive+'c')
                  if (exit_status==None or int(log(int(exit_status/256),2))>2):
                      drives.append(drive+'c')
              except :
                  continue
      except :
          verboselog('Failed to list OpenBSD disks')
  elif os.uname()[0]=="FreeBSD" :
      try :
          sysctl_kerndisks=os.popen('sysctl kern.disks')
          kerndisks=string.strip(sysctl_kerndisks.readline())
          for drive in string.split(kerndisks)[1:] :
              if drive[:2] in ['md','cd','fd'] : continue # Ignore Memory Disks, CD-ROM drives and Floppy
              try :
                  verboselog('Trying '+drive+'...')
                  exit_status=read_values(drive)
                  if (exit_status==None or int(log(int(exit_status/256),2))>2):
                      drives.append(drive)
              except :
                  continue
      except :
          verboselog('Failed to list FreeBSD disks')
  elif os.uname()[0]=="NetBSD" :
      try :
          sysctl_kerndisks=os.popen('sysctl hw.disknames')
          kerndisks=string.strip(sysctl_kerndisks.readline())
          for drive in string.split(kerndisks)[2:] :
              if drive[:2] in ['md','cd','fd'] : continue # Ignore Memory Disks, CD-ROM drives and Floppy
              try :
                  verboselog('Trying '+drive+'c...')
                  exit_status=read_values(drive+'c')
                  if (exit_status==None or int(log(int(exit_status/256),2))>2):
                      drives.append(drive+'c')
              except :
                  continue
      except :
          verboselog('Failed to list NetBSD disks')
  elif os.uname()[0]=="SunOS" :
      try :
          from glob import glob
          for drivepath in glob('/dev/rdsk/*s2') :
              try :
                  drive=os.path.basename(drivepath)
                  verboselog('Trying rdsk'+drive+'...')
                  exit_status=read_values('rdsk'+drive)
                  if (exit_status==None or int(log(int(exit_status/256),2))>2):
                      drives.append('rdsk'+drive)
              except :
                  continue
          for drivepath in glob('/dev/rmt/*') :
              try :
                  drive=os.path.basename(drivepath)
                  verboselog('Trying rmt'+drive+'...')
                  exit_status=read_values('rmt'+drive)
                  if (exit_status==None or int(log(int(exit_status/256),2))>2):
                      drives.append('rmt'+drive)
              except :
                  continue
      except :
          verboselog('Failed to list SunOS disks')
  return(drives)
  
def open_state_file(hard_drive,mode) :
    global statefiledir
    return open(statefiledir+'/smart-'+hard_drive+'.state',mode)

def update_state_file(hard_drive) :
    try:
        verboselog('Saving statefile')
        pickle.dump(smart_values,open_state_file(hard_drive,"w"))
    except :
        verboselog('Error trying to save state file! Check access rights')  
        
for drive in find_smart_drives():
  read_values(drive);
  update_state_file(drive);